/* logkf.c (C)2001 Tachibanamasashi <moomin@happymusic.com>
   Distribution, usage and modification of this program is permitted
   only GPL or LGPL. */

#include <stdlib.h>
#include <malloc.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "logkf.h"

int main (int argc, char *argv[]) {
  char sep[2], filepath[256], *ptr, filename[1024], linebuf[4096];
  char envlang, outcode, encode, verbose;
  FILE *fp;
  int i;

  /* detect kanji code in system environment */
  envlang = US_ASCII;
  strcpy(sep, "/");
#ifdef WIN32
  envlang = SJIS;
  strcpy(sep, "\\");
#endif
#ifdef DOS
  envlang = SJIS;
  strcpy(sep, "\\");
#endif
#ifdef MAC
  envlang = SJIS;
#endif
#ifdef UNIX
  if ((ptr = (char *)getenv("LC_ALL")) || (ptr = (char *)getenv("LC_CTYPE")) || (ptr = (char *)getenv("LANG"))) {
    if (strstr(ptr, "UTF-8")) {
      envlang = UTF8;
    } else if (strstr(ptr, "SJIS")) {
      envlang = SJIS;
    } else if (strstr(ptr, "ISO2022")) {
      envlang = JIS;
    } else {
      envlang = EUCJP;
    }
  } else {
    envlang = US_ASCII;
  }
#endif

  /* find out current file path when on run-time */
  if (argv[0] != NULL && strlen(argv[0]) < sizeof(filepath)) {
    strncpy(filepath, argv[0], sizeof(filepath)-1);
    if ((ptr = strstr(filepath, sep)) != NULL) {
      while (strcmp(ptr+1, "") && strstr(ptr+1, sep) != NULL) ptr = strstr(ptr+1, sep);
      strcpy(ptr+1, "");
    } else {
      strcpy(filepath, "");
    }
  } else {
    strcpy(filepath, "");
  }

  outcode = US_ASCII;
  encode = NOT_ASSIGNED_YET;
  verbose = ON;

  filename[0] = '\0';

  i = 1;
  while (i < argc && argv[i] != NULL && argv[i][0] == '-') {
    /* command line option switch */
    int col;

    /* show version of this program if added command line option '--version' */
    if (!strcmp(argv[i], "--version")) {
      if (envlang == US_ASCII) {
        fprintf(stdout, "%s version %s\n", program_name, program_version);
      } else {
        jp_fprintf(envlang, stdout, "%s \xA5\xD0\xA1\xBC\xA5\xB8\xA5\xE7\xA5\xF3 %s\n", program_name, program_version);
      }
      exit(0);
    }

    /* show version of this program if added command line option '--help' */
    if (!strcmp(argv[i], "--help")) {
      if (envlang == US_ASCII) {
        fprintf(stdout, "%s version %s\n", program_name, program_version);
        fputs(program_info, stdout);
        fprintf(stdout, "\n%s\n", program_usage);
      } else {
        jp_fprintf(envlang, stdout, "%s \xA5\xD0\xA1\xBC\xA5\xB8\xA5\xE7\xA5\xF3 %s\n", program_name, program_version);
        jp_fputs(envlang, program_info_jp, stdout);
        jp_fprintf(envlang, stdout, "\n%s\n", program_usage_jp);
      }
      exit(0);
    }

    /* discover other command line options */
    col = 1;
    while (argv[i][col] != '\0') {
      switch (argv[i][col]) {
        case 'e':
          if (outcode == US_ASCII) {
            outcode = UTF8;
          } else {
            if (verbose == ON) fprintf(stderr, "%s: Missing paramater Cannot open file \"%s\"\n", program_name, filename);
            exit(1);
          }
          break;
        case 's':
          outcode = SJIS;
          break;
        case 'j':
          outcode = JIS;
          break;
        case '8':
          outcode = UTF8;
          break;
        case 'u':
          encode = ON;
          break;
        case 'n':
          encode = OFF;
          break;
        case 'v':
          verbose = ON;
          break;
        case 'S':
          verbose = OFF;
      }
      col++;
    }
    i++;
  }

  /* if not set output kanji code in command line parameter, use environment settings */
  if (outcode == US_ASCII) {
    if (envlang != US_ASCII) {
      outcode = envlang;
    } else {
      outcode = EUCJP;
    }
  }

  /* if log-file name is set without directory path, add current file path */
  if (i < argc) {
    if (strstr(argv[1], sep) == NULL && strlen(filepath)+strlen(argv[i]) < sizeof(filename)) {
      sprintf(filename, "%s%s", filepath, argv[i]);
    } else {
      strncpy(filename, argv[i], sizeof(filename));
    }
    i++;
  }

  /* if target log-file name is set without directory path, add current file path */
  if (filename[0] == '\0') {
    if (envlang == US_ASCII) {
      fputs(program_info, stdout);
    } else {
      jp_fputs(envlang, program_info_jp, stdout);
    }
    exit(0);
  }

  /* main process */
  linebuf[sizeof(linebuf)-1] = '\0';
  if((fp = fopen(filename, "r")) != NULL){
    while (fgets(linebuf, sizeof(linebuf)-1, fp) != NULL) {
      /* read a line from target log-file, transform, and output! */
      convert_logtext(outcode, encode, linebuf, sizeof(linebuf));
      fprintf(stdout, "%s", linebuf);
    }
    fclose(fp);
  } else {
    if (verbose == ON) fprintf(stderr, "%s: Cannot open file \"%s\"\n", program_name, filename);
    exit(1);
  }

  return 0;
}

int convert_logtext(char outstyle, char encode, unsigned char *buf, int buf_size) {
  int i, j, k, max, space;
  unsigned char str, flg;

  /* in this function, convert japanese strings in text buffer to specified */
  /* code. first, transform all text to euc-jp code using decode_to_eucjp(). */
  /* next, if euc-jp code are not specified, call text convert routine. */
  /* if encode parameter is ON, do re-encode the converted characters. */
  /* note: japanese text are usually encoded in log-files as like '%A4%A2...' */

  if (0 < buf_size && 0 < (max = strlen(buf))) {
    space = buf_size - max-1;
    i = max;
    while (0 <= i) {
      buf[i + space] = buf[i];
      i--;
    }

    i = space;
    j = 0;
    max = strlen(buf);
    while (i < buf_size-1) {
      if (buf[i] == 0x20 || buf[i] == '+' || buf[i] == '&' || buf[i] == '"') {
        do {
          buf[j++] = buf[i++];
        } while (buf[i] == 0x20 || buf[i] == '+' || buf[i] == '&' || buf[i] == '"');
      } else {
        k = j;
        flg = 0;
        do {
          if (max <= i-space) break;
          buf[j] = buf[i++];
          if (flg == 0 && (buf[j] == '%' || (0x80 & buf[j]) || buf[j] == 0x1b)) flg = 1;
          j++;
        } while (buf[i] != 0x20 && buf[i] != '+' && buf[i] != '&' && buf[i] != '"');
        str = buf[i];
        if (flg == 1) {
          buf[j] = '\0';
          decode_to_eucjp(&buf[k]);

          switch (outstyle) {
            case SJIS:
              euc_to_sjis(&buf[k], i-k);
              break;
            case JIS:
              euc_to_jis(&buf[k], i-k);
              break;
            case UTF8:
              euc_to_utf8(&buf[k], i-k);
          }

          if (encode == ON) encode_logtext(&buf[k], i-k);
          j = k + strlen(&buf[k]);
        }
        if (str != '\0') {
          buf[j++] = str;
          i++;
        }
      }
    }
    if (j <= max) buf[j] = '\0';
  } else {
    return 0;
  }
  return strlen(buf);
}

char *decode_to_eucjp(unsigned char *buf) {
  int buf_size, ext_size, col, i, ioffset, imax, bad_euc, bad_sjis, bad_utf8, hankana;
  unsigned int c, d;
  enum {NORMAL, KANJI, HANKANA} mode = NORMAL;
  unsigned char a0, a1;

  buf_size = strlen(buf);

  ioffset = -1;
  i = col = 0;
  while (i < buf_size) {
    if (buf[i] == '%' && i+2 < buf_size && (('0' <= buf[i+1] && buf[i+1] <= '9') ||
      ('A' <= buf[i+1] && buf[i+1] <= 'F') || ('a' <= buf[i+1] && buf[i+1] <= 'f'))) {
      switch (buf[i + 1]) {
        case 'A':
        case 'a':
          a0 = 0xA0;
          break;
        case 'B':
        case 'b':
          a0 = 0xB0;
          break;
        case 'C':
        case 'c':
          a0 = 0xC0;
          break;
        case 'D':
        case 'd':
          a0 = 0xD0;
          break;
        case 'E':
        case 'e':
          a0 = 0xE0;
          break;
        case 'F':
        case 'f':
          a0 = 0xF0;
          break;
        default:
          a0 = (buf[i + 1] - '0') << 4;
      }
      switch (buf[i + 2]) {
        case 'A':
        case 'a':
          a1 = 0x0A;
          break;
        case 'B':
        case 'b':
          a1 = 0x0B;
          break;
        case 'C':
        case 'c':
          a1 = 0x0C;
          break;
        case 'D':
        case 'd':
          a1 = 0x0D;
          break;
        case 'E':
        case 'e':
          a1 = 0x0E;
          break;
        case 'F':
        case 'f':
          a1 = 0x0F;
          break;
        default:
          a1 = buf[i + 2] - '0';
      }
      buf[col] = (a0 | a1);
      if (0 < col && buf[col-1] == 0xa1 && buf[col] == 0xa1) {
        /* EUC 2byte-Space */
        buf[--col] = '+';
      }
      if (ioffset == -1 && ((0x80 & buf[col]) || buf[col] == 0x1b)) {
        ioffset = col;
      }
      col++;
      i += 3;
    } else {
      buf[col] = buf[i++];
      if (ioffset == -1 && buf[col] == '+') {
        ioffset = col;
      }
      col++;
    }
  }
  buf[col] = '\0';
  if (ioffset == -1) ioffset=0;

  imax = strlen(buf);
  ext_size = buf_size - imax;

  bad_euc = 0;
  for (i = ioffset; i < imax; i++) {
    if (iseuc(buf[i]) && (++i < imax)) {
      if (!iseuc(buf[i])) {
        bad_euc += 10;
        i--;
      } else if (0x50 <= buf[i-1]) {
        bad_euc++;
      }
    } else if ((buf[i] == 0x8e) && (++i < imax)) {
      if (ishankana(buf[i])) {
        bad_euc++;
      } else {
        bad_euc += 10;
        i--;
      }
    } else if (0x80 <= buf[i]) {
      bad_euc++;
    }
  }
  bad_sjis = 0;
  for (i = ioffset; i < imax; i++) {
    if (issjis1(buf[i]) && (++i < imax)) {
      if (!issjis2(buf[i])) {
        bad_sjis += 10;
        i--;
      } else if (0x989f <= (((unsigned int)buf[i-1] << 8) | (unsigned int)buf[i])) {
        bad_sjis++;
      }
    } else if (0x80 <= buf[i]) {
      if (ishankana(buf[i])) {
        bad_sjis++;
      } else {
        bad_sjis += 10;
      }
    }
  }
  if (i+2 < imax && buf[i-1] == 0xef && buf[i] == 0xbb && buf[i+1] == 0xbf) {
    bad_euc += 10;
    bad_sjis += 10;
  }
  bad_utf8 = 0;
  for (i = ioffset; i < imax; i++) {
    if (isutf8_1(buf[i]) && (++i < imax)) {
      if (isutf8_2(buf[i]) && (++i < imax)) {
        if (3 <= utf8_len(buf[i-2])) {
          if (!isutf8_3(buf[i])) {
            bad_utf8 += 10;
            i -= 2;
          }
        } else {
          i--;
        }
      } else {
        bad_utf8 += 5;
        i--;
      }
    } else if (iseuc(buf[i])) {
      bad_utf8 += 10;
    }
  }

  if (bad_sjis < bad_euc && bad_sjis <= bad_utf8) {
    /* JIS or SJIS to EUC */

    if (ext_size) for (i = imax; ioffset <= i; i--) buf[i + ext_size] = buf[i];

    mode = NORMAL;
    hankana = 0;
    i = ext_size + ioffset;
    col = ioffset;
    while (i < (ext_size + imax)) {
      if (buf[i] == 0x1b) {
        if (buf[i+1] == '$') {
          if (i+2 < (ext_size + imax) && (buf[i+2] == '@' || buf[i+2] == 'B')) {
            mode = KANJI;
            i += 2;
          } else {
            buf[col++] = 0x1b;
          }
        } else if (buf[i+1] == '(') {
          if (i+2 < (ext_size + imax) && (buf[i+2] == 'B' || buf[i+2] == 'J')) {
            mode = NORMAL;
            i += 2;
          } else {
            buf[col++] = 0x1b;
            i++;
          }
        } else {
            buf[col++] = 0x1b;
        }
      } else if (buf[i] == 0x0e) {
        mode = HANKANA;
      } else if (buf[i] == 0x0f) {
        mode = NORMAL;
      } else if ((mode == KANJI) && (isjis(buf[i]) && isjis(buf[i+1]))) {
        buf[col++] = (buf[i] | 0x80);
        buf[col++] = (buf[i+1] | 0x80);
        i++;
      } else if ((mode == HANKANA) && (0x20 <= buf[i] && buf[i] <= 0x5f)) {
        buf[col++] = 0x8e;
        buf[col++] = (buf[i] | 0x80);
      } else if (issjis1(c = buf[i]) && issjis2(d = buf[i+1])) {
        if (c <= 0x9f) {
          if (d < 0x9f) {
            c = (c << 1) - 0xe1;
          } else {
            c = (c << 1) - 0xe0;
          }
        } else {
          if (d < 0x9f) {
            c = (c << 1) - 0x161;
          } else {
            c = (c << 1) - 0x160;
          }
        }
        if (d < 0x7f) {
          d -= 0x1f;
        } else if (d < 0x9f) {
          d -= 0x20;
        } else {
          d -= 0x7e;
        }
        buf[col++] = (unsigned char)(c | 0x80);
        buf[col++] = (unsigned char)(d | 0x80);
        i++;
      } else if (ishankana(buf[i])) {
        if (++hankana < ext_size) {
          buf[col++] = 0x8e;
          buf[col++] = buf[i];
        } else {
          buf[col++] = '?';
        }
      } else {
          buf[col++] = buf[i];
      }
      if (0 <= col-2 && buf[col-2] == 0xa1 && buf[col-1] == 0xa1) {
        /* EUC 2byte-Space */
        buf[col-2] = '+';
        col--;
      }
      i++;
    }
    buf[col] = '\0';
  } else if (2 <= (imax - ioffset) && bad_utf8 < bad_euc && bad_utf8 < bad_sjis) {
    /* UTF-8 to EUC */

    if (ext_size) for (i = imax; ioffset <= i; i--) buf[i + ext_size] = buf[i];

    i = ext_size + ioffset;
    col = ioffset;
    while (i < (ext_size + imax)) {
      if (isutf8_1(buf[i]) && i+utf8_len(buf[i])-1 < (ext_size + imax)) {
        int pos, old_pos, min, max;

        if (buf[i] == 0xef && buf[i+1] == 0xbb && buf[i+2] == 0xbf) {
          i += 3;
          continue;
        }

        if (buf[i] == 0xc2 && buf[i+1] == 0xa5) {
          buf[i] = 0x5c;
          i += 2;
          continue;
        }

        if (buf[i] == 0xe2 && buf[i+1] == 0x80 && buf[i+2] == 0xbe) {
          buf[i] = 0x7e;
          i += 3;
          continue;
        }

        if (utf8_len(buf[i]) == 2) {
          a0 = 0x07 & (buf[i] >> 2);
          a1 = (buf[i] << 6) | (0x3f & buf[i+1]);
        } else {
          a0 = (buf[i] << 4) | (0x0f & (buf[i+1] >> 2));
          a1 = (buf[i+1] << 6) | (0x3f & buf[i+2]);
        }
 
        min = 0;
        max = UNICODE_EUCJP_TABLE_MAX-1;
        old_pos = -1;
        pos = (min+max)/2;
        do {
          if (unicode_eucjp[pos][0] == a0 && unicode_eucjp[pos][1] == a1) {
            buf[col++] = unicode_eucjp[pos][2];
            buf[col++] = unicode_eucjp[pos][3];
            i += (utf8_len(buf[i])-1);
            break;
          }

          old_pos = pos;
          if (unicode_eucjp[old_pos][0] < a0 || (unicode_eucjp[old_pos][0] == a0 && unicode_eucjp[old_pos][1] < a1)) {
            pos = (old_pos+1+max)/2;
            min = old_pos+1;
          } else {
            pos = (min+old_pos-1)/2;
            max = old_pos-1;
          }
        } while (pos != old_pos && min <= max);

        if (unicode_eucjp[pos][0] != a0 || unicode_eucjp[pos][1] != a1) buf[col++] = buf[i];
      } else {
        buf[col++] = buf[i];
      }
      if (1 < col && buf[col-2] == 0xa1 && buf[col-1] == 0xa1) {
        /* EUC 2byte-Space */
        buf[col-2] = '+';
        col--;
      }
      i++;
    }
    buf[col] = '\0';
  }

  return buf;
}

char *euc_to_sjis(unsigned char *buf, int buf_size) {
  /* EUC to SJIS */
  int i, col, len, ext_size;

  len = strlen(buf);
  ext_size = buf_size-len-1;

  if (0 < ext_size) {
    for (i = len; 0 <= i; i--) buf[i+ext_size] = buf[i];
  } else {
    ext_size = 0;
  }

  i = ext_size;
  col = 0;
  while (i-ext_size < len && col < buf_size) {
    int c, d;
    if (iseuc(c = (int)buf[i]) && iseuc(d = (int)buf[i+1])) {
      c &= 0x7f;
      d &= 0x7f;
      if (c & 1) {
        if (d < 0x60) {
          d += 0x1f;
        } else {
          d += 0x20;
        }
      } else {
         d += 0x7e;
      }
      if (c < 0x5f) {
        c = (c + 0xe1) >> 1;
      } else {
        c = (c + 0x161) >> 1;
      }
      buf[col++] = (unsigned char)c;
      buf[col++] = (unsigned char)d;
      i++;
    } else if ((buf[i] == 0x8e) && ishankana(buf[i+1])) {
      buf[col++] = buf[i+1];
      i++;
    } else {
      buf[col++] = buf[i];
    }
    i++;
  }
  buf[col] = '\0';
  return buf;
}

char *euc_to_jis(unsigned char *buf, int buf_size) {
  /* EUC to JIS */
  int i, col, len, ext_size;
  enum {NORMAL, KANJI, HANKANA} mode = NORMAL;

  len = strlen(buf);
  ext_size = buf_size-len-1;

  if (0 < ext_size) {
    for (i = len; 0 <= i; i--) buf[i+ext_size] = buf[i];
  } else {
    ext_size = 0;
  }

  i = ext_size;
  col = 0;
  while (i-ext_size < len && col < buf_size) {
    int c, d;
    if (iseuc(c = (int)(buf[i])) && iseuc(d = (int)(buf[i+1])) && col+1 < buf_size) {
      if (mode != KANJI && col+4 < buf_size) {
        if (mode == HANKANA && col+5 < buf_size) {
          buf[col++] = 0x0f;
        }
        mode = KANJI;
        buf[col++] = 0x1b;
        buf[col++] = '$';
        buf[col++] = 'B';
      }
      buf[col++] = (unsigned char)(c & 0x7f);
      buf[col++] = (unsigned char)(d & 0x7f);
      i++;
    } else if (c == 0x8e && col+1 < buf_size) {
      if (ishankana(d = (unsigned int)buf[i+1])) {
        if (mode != HANKANA && col+2 < buf_size) {
          if (mode == KANJI && col+5 < buf_size) {
            buf[col++] = 0x1b;
            buf[col++] = '(';
            buf[col++] = 'B';
          }
          mode = HANKANA;
        }
        buf[col++] = 0x0e;
        buf[col++] = (unsigned char)(d & 0x7f);
        i++;
      }
    } else {
      if (mode == KANJI && col+3 < buf_size) {
        buf[col++] = 0x1b;
        buf[col++] = '(';
        buf[col++] = 'B';
      } else if (mode == HANKANA && col+1 < buf_size) {
        buf[col++] = 0x0f;
      }
      mode = NORMAL;
      buf[col++] = (unsigned char)c;
    }
    i++;
  }
  if (mode == KANJI && col+2 < buf_size) {
    buf[col++] = 0x1b;
    buf[col++] = '(';
    buf[col++] = 'B';
  } else if (mode == HANKANA && col < buf_size) {
    buf[col++] = 0x0f;
  }
  buf[col] = '\0';
  return buf;
}

char *euc_to_utf8(unsigned char *buf, int buf_size) {
  /* EUC to UTF-8 */
  int i, col, len, ext_size;

  len = strlen(buf);
  ext_size = buf_size-len-1;

  if (0 < ext_size) {
    for (i = len; 0 <= i; i--) buf[i+ext_size] = buf[i];
  } else {
    ext_size = 0;
  }

  i = ext_size;
  col = 0;
  while (i-ext_size < len && col < buf_size) {
    unsigned char c, d, a0, a1;
    int l;
    if (iseuc(c = buf[i]) && iseuc(d = buf[i+1])) {
      for (l = 0; l < UNICODE_EUCJP_TABLE_MAX; l++) {
        if (unicode_eucjp[l][2] == c && unicode_eucjp[l][3] == d) break;
      }
      if (l < UNICODE_EUCJP_TABLE_MAX) {
        a0 = unicode_eucjp[l][0];
        a1 = unicode_eucjp[l][1];
        if (col+2 < buf_size) {
          if (a0 <= (unsigned char)0x07 && (unsigned char)0x80 <= a1) {
            buf[col++] = 0xc0 | (a1 >> 6);
            buf[col++] = 0x80 | (0x3f & a1);
          } else {
            buf[col++] = 0xe0 | (a0 >> 4);
            buf[col++] = 0x80 | ((0x3f & (a0 << 2)) | (a1 >> 6));
            buf[col++] = 0x80 | (0x3f & a1);
          }
          i++;
        } else {
          buf[col++] = c;
        }
      } else {
        buf[col++] = c;
      }
    } else {
      buf[col++] = c;
    }
    i++;
  }
  buf[col] = '\0';

  return buf;
}

int encode_logtext(unsigned char *buf, int buf_size) {
  int ext_size, i, j, imax;
  unsigned char a0, a1;

  imax = strlen(buf);
  ext_size = buf_size - imax;
  if (ext_size) {
    for (i = imax; 0 <= i; i--) buf[i + ext_size] = buf[i];

    j = 0;
    for (i = 0; i+j < buf_size && j <= ext_size; i++) {
      if ((buf[i + ext_size] & 0x80) || buf[i + ext_size] == '+' || buf[i + ext_size] == '"' || buf[i + ext_size] == '%' ||
          (buf[i + ext_size] <= 0x20 && buf[i + ext_size] != 0x0d && buf[i + ext_size] != 0x0a)) {
        if (buf[i + ext_size] == '\0') break;
        if (i+j+2 < buf_size) {
          a0 = ((buf[i + ext_size] & 0xf0) >> 4);
          a0 = (a0 < 10)?'0'+a0:'A'-10+a0;
          a1 = (buf[i + ext_size] & 0x0f);
          a1 = (a1 < 10)?'0'+a1:'A'-10+a1;
          buf[i+j] = '%';
          buf[i+j+1] = a0;
          buf[i+j+2] = a1;
          j += 2;
        } else {
          buf[i+j] = buf[i + ext_size];
        }
      } else {
        buf[i+j] = buf[i + ext_size];
      }
    }
    buf[i+j] = '\0';
  }

  return strlen(buf);
}

int jp_fputs(char outstyle, char *str, FILE *outf) {
  int buf_size, ret;
  char *str_buf;

  if (outf != NULL) {
    str_buf = (char *)malloc((buf_size = strlen(str)*3+100)+1);
    if (str_buf == (char *)NULL) return (int)NULL;
    strcpy(str_buf, str);

    switch (outstyle) {
      case SJIS:
        ret = fputs(euc_to_sjis(str_buf, buf_size), outf);
        break;
      case JIS:
        ret = fputs(euc_to_jis(str_buf, buf_size), outf);
        break;
      case UTF8:
        ret = fputs(euc_to_utf8(str_buf, buf_size), outf);
        break;
      default:
        ret = fputs(str_buf, outf);
    }

    free(str_buf);
  } else {
    ret = fputs(NULL, outf);
  }
  return ret;
}

int jp_fprintf(char outstyle, FILE *outf, char *format, ...) {
  int format_len, buf_size, col, ret;
  char *str_buf;
  va_list argp;
  char *p;

  if (format != NULL) {
    ret = 0;
    format_len = strlen(format);
    va_start(argp, format);
    col = 0;
    while (col < format_len) {
      int size;
      size = strcspn(&format[col], "%%");
      if (0 < size) {
        str_buf = (char *)malloc((buf_size = size*3+100)+1);
        if (str_buf == (char *)NULL) return (int)NULL; /* error ! */
        strncpy(str_buf, &format[col], size);
        str_buf[size] = '\0';
        switch (outstyle) {
          case SJIS:
            euc_to_sjis(str_buf, buf_size);
            break;
          case JIS:
            euc_to_jis(str_buf, buf_size);
            break;
          case UTF8:
            euc_to_utf8(str_buf, buf_size);
        }
        ret += fprintf(outf, "%s", str_buf);
        free(str_buf);
        col += size-1;
      } else if (format[col] == '%'){
        if (format_len <= col+1) {
            ret += fprintf(outf, "%%");
        } else switch (format[col+1]) {
          case '%':
            ret += fprintf(outf, "%%");
            col++;
            break;
          case 's':
            p = va_arg(argp, char *);
            size = strlen((char *)p);
            if (0 < size) {
              str_buf = (char *)malloc((buf_size = size*3+100)+1);
              if (str_buf == (char *)NULL) return (int)NULL; /* error ! */
              strcpy(str_buf, p);
              switch (outstyle) {
                case SJIS:
                  euc_to_sjis(str_buf, buf_size);
                  break;
                case JIS:
                  euc_to_jis(str_buf, buf_size);
                  break;
                case UTF8:
                  euc_to_utf8(str_buf, buf_size);
              }
              ret += fprintf(outf, "%s", str_buf);
              free(str_buf);
            }
            col++;
            break;
          case '*':
            if (format[col+2] == 's') {
              int l;
              l = (int)va_arg(argp, int *);
              p = va_arg(argp, char *);
              size = strlen((char *)p);
              if (0 < size) {
                str_buf = (char *)malloc((buf_size = size*3+100)+1);
                if (str_buf == (char *)NULL) return (int)NULL; /* error ! */
                strcpy(str_buf, p);
                switch (outstyle) {
                  case SJIS:
                    euc_to_sjis(str_buf, buf_size);
                    break;
                  case JIS:
                    euc_to_jis(str_buf, buf_size);
                    break;
                  case UTF8:
                    euc_to_utf8(str_buf, buf_size);
                }
                ret += fprintf(outf, "%*s", l, str_buf);
                free(str_buf);
              }
              col += 2;
              break;
            }
          default:
            size = strcspn(&format[col], "abcdefgijkmnopqrstuvwxyzABCDEFGHIJKMNOPQRSTUVWXYZ")+1;
            if (1 < size) {
              str_buf = (char *)malloc((buf_size = size*3+100)+1);
              if (str_buf == (char *)NULL) return (int)NULL; /* error ! */
              strncpy(str_buf, &format[col], size);
              str_buf[size] = '\0';
              switch (outstyle) {
                case SJIS:
                  euc_to_sjis(str_buf, buf_size);
                  break;
                case JIS:
                  euc_to_jis(str_buf, buf_size);
                  break;
                case UTF8:
                  euc_to_utf8(str_buf, buf_size);
              }
              ret += fprintf(outf, str_buf, va_arg(argp, void *));
              free(str_buf);
              col += size-1;
            }
        }
      }
      col++;
    }
    va_end(argp);
  } else {
    ret = fprintf(outf, NULL);
  }
  return ret;
}
